# IPToCountry release 2 by David Bowland
# ./addons/eventscripts/_libs/python/iptocountry.py
# Database provided by http://Software77.net

"""
To use this library, you must first import it:

import iptocountry


The following functions are available:

iptocountry.get_country(ip)
   - Returns full the country name, three-letter country abbreviation corresponding to IP
   ip = IP address to resolve to country name


iptocountry.get_location_data(ip):
   - Returns a dictionary of location data corresponding to IP
      * Dictionary keys:
         ip_from      = Bottom long IP of location range
         ip_to        = Upper long IP of location range
         registry     = International registry agency through which the IP is registered
         assigned     = Seconds since epoch the IP was assigned
         country_2    = Two-character international country code
         country_3    = Three-character international country code (preferred because more unique)
         country_long = Country name
   ip = IP address to resolve to location data


iptocountry.update_data(url=None):
   - Syncs server database with internet data--VERY SLOW
   - Returns True if update was sucessful otherwise returns False
   url = URL to download IPToCountry database


iptocountry.get_last_update():
   - Returns the time of the last database update (in seconds since epoch)


The following instance of CountryResolve is publicly exposed:

iptocountry.service


The database for this library is provided by http://Software77.net. For more information, please see:

http://software77.net/cgi-bin/ip-country/geo-ip.pl

NO CLAIMS ARE MADE REGARDING THE ACCURACY OF THIS DATABASE. Please use this library with that in mind.
"""


import os
import StringIO
import sys
import traceback
import urllib
import zipfile


try:
   import psyco
   psyco.full()

   # Source server running EventScripts
   import es
   dbgmsg   = es.dbgmsg
   lib_path = es.getAddonPath('_libs') + '/python/'

except ImportError:
   # IDLE
   def dbgmsg(x, msg): print msg
   lib_path = os.getcwd() + '/'


class CountryResolve:
   def __init__(self, url=None):
      """Loads database for later query"""
      self.ip_data   = []
      self.base_path = lib_path

      # Loads database
      if self.get_last_update():
         self._load_data()

      else:
         dbgmsg(0, 'No IPToCountry database found! Downloading database.')
         if not self.update_data(url):
            raise IOError, 'Could not download IPToCountry database!'

   def get_location_data(self, ip):
      """
      Returns a dictionary of location data corresponding to IP
      Dictionary keys:
         ip_from      = Bottom long IP of location range
         ip_to        = Upper long IP of location range
         registry     = International registry agency through which the IP is registered
         assigned     = Seconds since epoch the IP was assigned
         country_2    = Two-character international country code
         country_3    = Three-character international country code (preferred because more unique)
         country_long = Country name
      More information can be found here: http://software77.net/cgi-bin/ip-country/geo-ip.pl
      """
      if not self.ip_data:
         # No database to query
         return {}

      # Formats the IP without port to a list
      if ':' in ip:
         ip = ip[:ip.find(':')]
      ip_list = map(int, ip.split('.'))

      # Validates the address for conversion to long format
      if len(ip_list) <> 4:
         raise ValueError, 'Invalid IP address \"%s\"' % ip

      # Converts the IP to long IP format
      long_ip = ip_list[3] + ip_list[2] * 256 + ip_list[1] * 65536 + ip_list[0] * 16777216

      # Finds the IP in the database
      for data in self.ip_data:
         if long_ip <= data['ip_to']:
            return data

      # No data found
      return {}

   def get_country(self, ip):
      """Returns full country name and three-letter country abbreviation corresponding to IP"""
      data = self.get_location_data(ip)

      if data:
         # Country data found
         return data['country_long'], data['country_3']

      else:
         # Country unknown
         return 'Unknown', 'Unknown'

   def update_data(self, url=None):
      """
      Syncs server database with internet data--VERY SLOW
      Returns True if update is sucessful, otherwise returns False
      """
      if not url:
         url = 'http://software77.net/cgi-bin/ip-country/geo-ip.pl?action=downloadZ'

      # The database is not yet updated
      return_val = False

      # Download database
      raw_zip = self._download_database(url)

      # If download was sucessful, parse the databse
      if raw_zip:
         return_val = self._parse_zip(raw_zip)

         # Close the downloaded zip file
         raw_zip.close()

      # Update complete
      return return_val

   def get_last_update(self):
      """Returns the time of the last database update (in seconds since epoch)"""
      path = self.base_path + 'iptocountry.db'
      return os.stat(path).st_mtime if os.path.isfile(path) else 0

   def _load_data(self):
      """
      Internal use recommended: This function is SLOW
      Loads the IP data into memory
      """
      self.ip_data[:] = []
      path = self.base_path + 'iptocountry.db'

      if os.path.isfile(path):
         data_db = open(path)

         # Reads the contents of the database to a dictionary
         for line in data_db.readlines():
            data = line.strip().split(',')
            self.ip_data.append({'ip_from':float(data[0]), 'ip_to':float(data[1]), 'registry':data[2], 'assigned':float(data[3]), 'country_2':data[4], 'country_3':data[5].strip(), 'country_long':data[6]})

         data_db.close()

      else:
         self.update_data()

   def _download_database(self, url):
      try:
         url        = urllib.urlopen(url)
         return_val = StringIO.StringIO(url.read())

         url.close()

      except:
         self.__show_exception(sys.exc_info())
         return_val = None

      return return_val

   def _parse_zip(self, raw_zip):
      try:
         zip = zipfile.ZipFile(raw_zip)

         filelist = map(lambda x: x.filename, zip.filelist)
         db_file  = 'IpToCountry.csv' if 'IpToCountry.csv' in filelist else filelist[0]

         raw_database = StringIO.StringIO(zip.read(db_file))
         return_val   = self.___parse_database(raw_database)

         raw_database.close()

         if return_val:
            self._load_data()

      except:
         self.__show_exception(sys.exc_info())
         return_val = False

      return return_val

   def ___parse_database(self, raw_database):
      """
      Internal use highly recommended
      Converts raw internet data into plain-text format
      """
      database = []

      for line in raw_database.readlines():
         if line.startswith('#'): continue

         if line.count(',') >= 6:
            database.append(line.strip().replace('\"', '') + '\n')

      if not database: return False

      # Saves the parsed database
      new_db = open(self.base_path + 'iptocountry.db', 'w')

      # Sorts the database for quicker reference
      new_db.writelines(sorted(database, key=lambda x: float(x[:x.find(',')])))
      new_db.close()

      return True

   def __show_exception(self, exc_info):
      dbgmsg(1, 'IPToCountry update exception:')
      for line in traceback.format_exception(*exc_info):
         dbgmsg(1, line[:~0])


# Public instance of CountryResolve
service = CountryResolve()


# Easy-reference functions to the public instance of CountryResolve
def get_country(ip):
   """Returns full country name and three-letter country abbreviation of IP"""
   return service.get_country(ip)
get_country.__doc__ = service.get_country.__doc__


def get_location_data(ip):
   """Returns a dictionary of location data of IP"""
   return service.get_location_data(ip)
get_location_data.__doc__ = service.get_location_data.__doc__


def update_data(url=None):
   """Syncs server database with internet data--VERY SLOW"""
   return service.update_data(url)
update_data.__doc__ = service.update_data.__doc__


def get_last_update():
   """Returns the time of the last database update (in seconds since epoch)"""
   return service.get_last_update()
get_last_update.__doc__ = service.get_last_update.__doc__